import os
import unittest

from dataset_creator.dataset import Dataset

from .data import test_data

FASTA_DATA_PATH = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'Fasta')


class TestFasta(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_dataset(self):
        dataset = Dataset(test_data, format='FASTA', partitioning='by gene')
        result = dataset.dataset_str
        expected = open(os.path.join(FASTA_DATA_PATH, 'dataset.fas')).read()
        self.assertEqual(expected, result)

    def test_partitioned_by_1st2nd_3rd(self):
        dataset = Dataset(test_data, format='FASTA', partitioning='1st-2nd, 3rd')
        result = dataset.dataset_str
        expected = open(os.path.join(FASTA_DATA_PATH, 'dataset_1st2nd_3rd.fas')).read()
        self.assertEqual(expected, result)

    def test_partitioned_by_1st2nd_3rd_only_1st2nd_codon_positions(self):
        dataset = Dataset(test_data, format='FASTA', partitioning='1st-2nd, 3rd',
                          codon_positions='1st-2nd')
        result = dataset.dataset_str
        expected = open(os.path.join(FASTA_DATA_PATH, 'dataset_1st2nd.fas')).read()
        self.assertEqual(expected, result)

    def test_partitioned_by_1st2nd_3rd_only_1st_codon_position(self):
        dataset = Dataset(test_data, format='FASTA', partitioning='1st-2nd, 3rd',
                          codon_positions='1st')
        result = dataset.dataset_str
        expected = open(os.path.join(FASTA_DATA_PATH, 'dataset_1st.fas')).read()
        self.assertEqual(expected, result)

    def test_partitioned_by_1st2nd_3rd_only_2nd_codon_position(self):
        dataset = Dataset(test_data, format='FASTA', partitioning='1st-2nd, 3rd',
                          codon_positions='2nd')
        result = dataset.dataset_str
        expected = open(os.path.join(FASTA_DATA_PATH, 'dataset_2nd.fas')).read()
        self.assertEqual(expected, result)

    def test_partitioned_by_1st2nd_3rd_only_3rd_codon_position(self):
        dataset = Dataset(test_data, format='FASTA', partitioning='1st-2nd, 3rd',
                          codon_positions='3rd')
        result = dataset.dataset_str
        expected = open(os.path.join(FASTA_DATA_PATH, 'dataset_3rd.fas')).read()
        self.assertEqual(expected, result)



from .utils import get_seq
from .base_dataset import DatasetBlock


class GenBankFASTADatasetBlock(DatasetBlock):
    def convert_to_string(self, block):
        """
        Takes a list of SeqRecordExpanded objects corresponding to a gene_code
        and produces the gene_block as string.

        :param block:
        :return: str.
        """
        out = ""
        for seq_record in block:
            taxon_id = '>{0}_{1}_{2} [org={0} {1}] [Specimen-voucher={2}] [note={3} gene, partial cds.] [Lineage=]'.format(
                seq_record.taxonomy['genus'],
                seq_record.taxonomy['species'],
                seq_record.voucher_code,
                seq_record.gene_code,
            )
            sequence = get_seq(seq_record, self.codon_positions, self.aminoacids,
                               self.degenerate)
            seq = sequence.seq
            if sequence.warning:
                self.warnings.append(sequence.warning)

            n = 60
            seq = [seq[i:i + n] for i in range(0, len(seq), n)]
            out += '{0}\n{1}\n'.format(taxon_id, "\n".join(seq))
        return out



import unittest

from seqrecord_expanded import SeqRecordExpanded

from dataset_creator.utils import get_seq
from dataset_creator.exceptions import WrongParameterFormat


class TestUtils(unittest.TestCase):
    def setUp(self):
        self.seq_record = SeqRecordExpanded('ATACGGTAG', table=1, reading_frame=1,
                                            voucher_code="CP100-10", gene_code="wingless")

    def test_get_seq(self):
        self.assertRaises(WrongParameterFormat, get_seq, seq_record=self.seq_record,
                          codon_positions='1st-3rd')

    def test_get_seq_sequence(self):
        """Test the returned sequence and warning"""
        result = get_seq(self.seq_record, codon_positions='ALL', aminoacids=True)
        self.assertEqual("IR*", result.seq)
        self.assertEqual("Gene wingless, sequence CP100-10 contains stop codons '*'",
                         result.warning)



#! encoding: utf-8
from collections import namedtuple
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

from .creator import Creator
from .utils import get_seq


class Dataset(object):
    """User's class for making datasets of several formats. It needs as input
    a list of SeqRecord-expanded objects with as much info as possible:

    Parameters:
        seq_records (list):     SeqRecordExpanded objects. The list should be
                                sorted by gene_code and then voucher code.
        format (str):           NEXUS, PHYLIP, TNT, MEGA, GenBankFASTA.
        partitioning (str):     Partitioning scheme: ``by gene`` (default),
                                ``by codon position`` (each) and ``1st-2nd, 3rd``.
        codon_positions (str):  Can be ``1st``, ``2nd``, ``3rd``, ``1st-2nd``,
                                ``ALL`` (default).
        aminoacids (boolean):   Returns the dataset as aminoacid sequences.
        degenerate (str):       Method to degenerate nucleotide sequences,
                                following Zwick et al. Can be ``S``, ``Z``,
                                ``SZ`` and ``normal``.
        outgroup (str):         voucher code to be used as outgroup for NEXUS
                                and TNT files.

    Attributes:
         _gene_codes_and_lengths (dict):   in the form ``gene_code: list``
                                           The list contains sequence lengths for its
                                           sequences. We assume the longest to be the
                                           real gene_code sequence length.

    Example:

        >>> dataset = Dataset(seq_records, format='NEXUS', codon_positions='1st',
        ...                   partitioning='by gene',
        ...                   )
        >>> print(dataset.dataset_str)
        '#NEXUS
        blah blah
        '
        >>> dataset = Dataset(seq_records, format='PHYLIP', codon_positions='ALL',
        ...                   partitioning='by gene',
        ...                   )
        >>> print(dataset.dataset_str)
        '100 10
        blah blah
        '
    """
    def __init__(self, seq_records, format=None, partitioning=None,
                 codon_positions=None, aminoacids=None, degenerate=None,
                 outgroup=None):
        self.warnings = []
        self.seq_records = self.sort_seq_records(seq_records)
        self.gene_codes = None
        self.number_taxa = None
        self.number_chars = None
        self.reading_frames = {}

        self.format = format
        self.partitioning = partitioning
        self.codon_positions = codon_positions
        self.aminoacids = aminoacids
        self.degenerate = degenerate
        self.outgroup = None

        self._validate_codon_positions(codon_positions)
        self._validate_partitioning(partitioning)
        self._validate_outgroup(outgroup)

        self.data = None
        self._gene_codes_and_lengths = OrderedDict()
        self._prepare_data()
        self.extra_dataset_str = None
        self.dataset_str = self._create_dataset()

    def sort_seq_records(self, seq_records):
        """Makes sure that SeqExpandedRecords are sorted by gene_code and then
        by voucher code.
        """
        unsorted_gene_codes = set([i.gene_code for i in seq_records])
        sorted_gene_codes = list(unsorted_gene_codes)
        sorted_gene_codes.sort(key=lambda x: x.lower())

        unsorted_voucher_codes = set([i.voucher_code for i in seq_records])
        sorted_voucher_codes = list(unsorted_voucher_codes)
        sorted_voucher_codes.sort(key=lambda x: x.lower())

        if sorted_gene_codes == list(unsorted_gene_codes) and \
                sorted_voucher_codes == list(unsorted_voucher_codes):
            return seq_records
        else:
            sorted_seq_records = []
            append = sorted_seq_records.append
            for gene_code in sorted_gene_codes:
                for voucher_code in sorted_voucher_codes:
                    for seq_record in seq_records:
                        if seq_record.gene_code == gene_code and \
                                seq_record.voucher_code == voucher_code:
                            append(seq_record)
            return sorted_seq_records

    def _validate_partitioning(self, partitioning):
        if partitioning is None:
            self.partitioning = 'by gene'
        elif partitioning not in ['by gene', 'by codon position', '1st-2nd, 3rd']:
            raise AttributeError("Partitioning parameter should be one of these: "
                                 "None, 'by gene', 'by codon position', '1st-2nd, 3rd")
        elif partitioning in ['by codon position', '1st-2nd, 3rd'] \
                and self.degenerate:
            raise ValueError("Cannot degenerate if partitions scheme is {0!r}".format(
                partitioning))
        elif partitioning in ['by codon position', '1st-2nd, 3rd'] and self.format == 'MEGA':
            raise ValueError("Cannot produce MEGA dataset with codon positions in different partitions")

    def _validate_codon_positions(self, codon_positions):
        if codon_positions is None:
            self.codon_positions = 'ALL'
        elif codon_positions not in ['1st', '2nd', '3rd', '1st-2nd', 'ALL']:
            raise AttributeError("Codon positions parameter should be one of these: "
                                 "None, '1st', '2nd', '3rd', '1st-2nd', 'ALL'")

    def _validate_outgroup(self, outgroup):
        if outgroup:
            good_outgroup = False
            for seq_record in self.seq_records:
                if seq_record.voucher_code == outgroup:
                    good_outgroup = True
                    break
            if good_outgroup:
                self.outgroup = outgroup
            else:
                raise ValueError("The given outgroup {0!r} cannot be found in the"
                                 "input sequence records.")
        else:
            self.outgroup = None

    def _prepare_data(self):
        """
        Creates named tuple with info needed to create a dataset.

        :return: named tuple
        """
        self._extract_genes()
        self._extract_total_number_of_chars()
        self._extract_number_of_taxa()
        self._extract_reading_frames()

        Data = namedtuple('Data', ['gene_codes', 'number_taxa', 'number_chars',
                                   'seq_records', 'gene_codes_and_lengths',
                                   'reading_frames'])
        self.data = Data(self.gene_codes, self.number_taxa, self.number_chars,
                         self.seq_records, self._gene_codes_and_lengths,
                         self.reading_frames)

    def _extract_genes(self):
        gene_codes = [i.gene_code for i in self.seq_records]
        unique_gene_codes = list(set(gene_codes))
        # this is better: unique_gene_codes.sort(key=str.lower)
        # but will not work in python2
        unique_gene_codes.sort(key=lambda x: x.lower())
        self.gene_codes = unique_gene_codes

    def _extract_total_number_of_chars(self):
        """
        sets `self.number_chars` to the number of characters as string.
        """
        self._get_gene_codes_and_seq_lengths()

        sum = 0
        for seq_length in self._gene_codes_and_lengths.values():
            sum += sorted(seq_length, reverse=True)[0]
        self.number_chars = str(sum)

    def _get_gene_codes_and_seq_lengths(self):
        for seq_record in self.seq_records:
            if seq_record.gene_code not in self._gene_codes_and_lengths:
                self._gene_codes_and_lengths[seq_record.gene_code] = []

            if self.aminoacids:
                seq = seq_record.translate()
            elif not self.aminoacids and self.degenerate is not None:
                seq = seq_record.degenerate(method=self.degenerate)
            else:
                sequence = get_seq(seq_record, self.codon_positions)
                seq = sequence.seq
                if sequence.warning:
                    self.warnings.append(sequence.warning)
            self._gene_codes_and_lengths[seq_record.gene_code].append(len(seq))

    def _extract_number_of_taxa(self):
        """
        sets `self.number_taxa` to the number of taxa as string
        """
        n_taxa = dict()
        for i in self.seq_records:
            if i.gene_code not in n_taxa:
                n_taxa[i.gene_code] = 0
            n_taxa[i.gene_code] += 1
        number_taxa = sorted([i for i in n_taxa.values()], reverse=True)[0]
        self.number_taxa = str(number_taxa)

    def _extract_reading_frames(self):
        for seq_record in self.seq_records:
            if seq_record.gene_code not in self.reading_frames:
                self.reading_frames[seq_record.gene_code] = seq_record.reading_frame

    def _create_dataset(self):
        creator = Creator(self.data, format=self.format,
                          codon_positions=self.codon_positions,
                          partitioning=self.partitioning,
                          aminoacids=self.aminoacids,
                          degenerate=self.degenerate,
                          outgroup=self.outgroup,
                          )
        self.warnings = creator.warnings
        self.extra_dataset_str = creator.extra_dataset_str
        dataset_str = creator.dataset_str

        return dataset_str



import os
import unittest

from dataset_creator.dataset import Dataset

from .data import test_data

MEGA_DATA_PATH = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'Mega')


class TestMega(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_dataset(self):
        dataset = Dataset(test_data, format='MEGA', partitioning='by gene')
        result = dataset.dataset_str
        expected = open(os.path.join(MEGA_DATA_PATH, 'dataset.meg')).read()
        self.assertEqual(expected, result)

    def test_dataset_as_aminoacids(self):
        dataset = Dataset(test_data, format='MEGA', partitioning='by gene',
                          aminoacids=True)
        result = dataset.dataset_str
        expected = open(os.path.join(MEGA_DATA_PATH, 'dataset_aa.meg')).read()
        self.assertEqual(expected, result)

    def test_dataset_as_degenerate(self):
        dataset = Dataset(test_data, format='MEGA', partitioning='by gene',
                          degenerate='S')
        result = dataset.dataset_str
        expected = open(os.path.join(MEGA_DATA_PATH, 'dataset_degenerate.meg')).read()
        self.assertEqual(expected, result)

    def test_dataset_as_degenerate_bad_partitioning_scheme(self):
        self.assertRaises(ValueError, Dataset, test_data, format='MEGA',
                          partitioning='1st-2nd, 3rd', degenerate='S')

    def test_partitioning_dataset_in_mega_format(self):
        self.assertRaises(ValueError, Dataset, test_data, format='MEGA',
                          partitioning='1st-2nd, 3rd')



# -*- coding: UTF-8 -*-
import six
if six.PY2:
    from StringIO import StringIO
else:
    from io import StringIO

import os
from collections import namedtuple
import uuid

from Bio import AlignIO

from .exceptions import WrongParameterFormat


def get_seq(seq_record, codon_positions, aminoacids=False, degenerate=None):
    """
    Checks parameters such as codon_positions, aminoacids... to return the
    required sequence as string.

    Parameters:
        seq_record (SeqRecordExpanded object):
        codon_positions (str):
        aminoacids (boolean):

    Returns:
        Namedtuple containing ``seq (str)`` and ``warning (str)``.
    """
    Sequence = namedtuple('Sequence', ['seq', 'warning'])

    if codon_positions not in [None, '1st', '2nd', '3rd', '1st-2nd', 'ALL']:
        raise WrongParameterFormat("`codon_positions` argument should be any of the following"
                                   ": 1st, 2nd, 3rd, 1st-2nd or ALL")
    if aminoacids:
        aa = seq_record.translate()
        if '*' in aa:
            warning = "Gene {0}, sequence {1} contains stop codons '*'".format(seq_record.gene_code,
                                                                               seq_record.voucher_code)
        else:
            warning = None
        return Sequence(seq=aa, warning=warning)

    if degenerate:
        return Sequence(seq=seq_record.degenerate(degenerate), warning=None)

    if codon_positions == '1st':
        return Sequence(seq=seq_record.first_codon_position(), warning=None)
    elif codon_positions == '2nd':
        return Sequence(seq=seq_record.second_codon_position(), warning=None)
    elif codon_positions == '3rd':
        return Sequence(seq=seq_record.third_codon_position(), warning=None)
    elif codon_positions == '1st-2nd':
        return Sequence(seq=seq_record.first_and_second_codon_positions(), warning=None)
    else:  # None and ALL
        return Sequence(seq=str(seq_record.seq), warning=None)


def convert_nexus_to_format(dataset_as_nexus, dataset_format):
    """
    Converts nexus format to Phylip and Fasta using Biopython tools.

    :param dataset_as_nexus:
    :param dataset_format:
    :return:
    """
    fake_handle = StringIO(dataset_as_nexus.replace('-', '_'))
    nexus_al = AlignIO.parse(fake_handle, 'nexus')
    tmp_file = make_random_filename()
    AlignIO.write(nexus_al, tmp_file, dataset_format)
    dataset_as_fasta = read_and_delete_tmp_file(tmp_file)
    return dataset_as_fasta


def make_random_filename():
    return '{0}.txt'.format(uuid.uuid4().hex)


def read_and_delete_tmp_file(filename):
    with open(filename, "r") as handle:
        contents = handle.read()

    if os.path.isfile(filename):
        os.remove(filename)

    return contents


def make_dataset_header(data, file_format, aminoacids):
    """Creates the dataset header for NEXUS files from ``#NEXUS`` to ``MATRIX``.

    Parameters:
        data (namedtuple):    with necessary info for dataset creation.
        file_format (str):    TNT, PHYLIP, NEXUS, FASTA
        aminoacids (boolean): If ``aminoacids is True`` the header will show
                              ``DATATYPE=PROTEIN`` otherwise it will be ``DNA``.
    """
    if aminoacids:
        datatype = 'PROTEIN'
    else:
        datatype = 'DNA'

    if file_format in ['NEXUS', 'PHYLIP', 'FASTA']:
        header = """
#NEXUS

BEGIN DATA;
DIMENSIONS NTAX={0} NCHAR={1};
FORMAT INTERLEAVE DATATYPE={2} MISSING=? GAP=-;
MATRIX
""".format(data.number_taxa, data.number_chars, datatype)

    elif file_format == 'MEGA':
        return "#MEGA\n!TITLE title;"

    else:  # file_format: TNT
        if aminoacids:
            molecule_type = "prot"
        else:
            molecule_type = "dna"
        header = """
nstates {0};
xread
{1} {2}""".format(molecule_type, data.number_chars, data.number_taxa)

    return header.strip()



import re
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

from .utils import get_seq


class DatasetBlock(object):
    """
    By default, the data sequences block generated is NEXUS and we use BioPython
    tools to convert it to other formats such as FASTA.
    However, sometimes the blo

    Parameters:
        data (named tuple):      containing:
                                  * gene_codes: list
                                  * number_chars: string
                                  * number_taxa: string
                                  * seq_records: list of SeqRecordExpanded objects
                                  * gene_codes_and_lengths: OrderedDict
        codon_positions (str):   str. Can be 1st, 2nd, 3rd, 1st-2nd, ALL (default).
        partitioning (str):
        aminoacids (boolean):
        degenerate (str):
        format (str):       NEXUS, PHYLIP or FASTA.
        outgroup (str):     Specimen code of taxon that should be used as outgroup.
    """
    def __init__(self, data, codon_positions, partitioning, aminoacids=None,
                 degenerate=None, format=None, outgroup=None):
        self.warnings = []
        self.data = data
        self.codon_positions = codon_positions
        self.partitioning = partitioning
        self.aminoacids = aminoacids
        self.degenerate = degenerate
        self.format = format
        self.outgroup = outgroup
        self._blocks = []

    def dataset_block(self):
        """Creates the block with taxon names and their sequences.

        Override this function if the dataset block needs to be different
        due to file format.

        Example:

            CP100_10_Aus_aus   ACGATRGACGATRA...
            CP100_11_Aus_bus   ACGATRGACGATRA...
            ...

        """
        self.split_data()
        out = []
        for block in self._blocks:
            out.append(self.convert_to_string(block))
        return '\n'.join(out).strip() + '\n;\nEND;'

    def split_data(self):
        """Splits the list of SeqRecordExpanded objects into lists, which are
        kept into a bigger list.

        If the file_format is Nexus, then it is only partitioned by gene. If it
        is FASTA, then it needs partitioning by codon positions if required.

        Example:

            >>> blocks = [
            ...     [SeqRecord1, SeqRecord2],  # for gene 1
            ...     [SeqRecord1, SeqRecord2],  # for gene 2
            ...     [SeqRecord1, SeqRecord2],  # for gene 3
            ...     [SeqRecord1, SeqRecord2],  # for gene 4
            ... ]
        """
        this_gene_code = None
        for seq_record in self.data.seq_records:
            if this_gene_code is None or this_gene_code != seq_record.gene_code:
                this_gene_code = seq_record.gene_code
                self._blocks.append([])
            list_length = len(self._blocks)
            self._blocks[list_length - 1].append(seq_record)

    def convert_to_string(self, block):
        """Takes a list of SeqRecordExpanded objects corresponding to a gene_code
        and produces the gene_block as string.

        Override this function if the dataset block needs to be different
        due to file format.

        This block will need to be split further if the dataset is FASTA or
        TNT and the partitioning scheme is 1st-2nd, 3rd.

        As the dataset is split into several blocks due to 1st-2nd, 3rd
        we cannot translate to aminoacids or degenerate the sequences.
        """
        if self.partitioning != '1st-2nd, 3rd':
            return self.make_datablock_by_gene(block)
        else:
            if self.format == 'FASTA':
                return self.make_datablock_considering_codon_positions_as_fasta_format(block)
            else:
                return self.make_datablock_by_gene(block)

    def make_datablock_considering_codon_positions_as_fasta_format(self, block):
        block_1st2nd = OrderedDict()
        block_1st = OrderedDict()
        block_2nd = OrderedDict()
        block_3rd = OrderedDict()

        for seq_record in block:  # splitting each block in two
            if seq_record.gene_code not in block_1st2nd:
                block_1st2nd[seq_record.gene_code] = []
            if seq_record.gene_code not in block_1st:
                block_1st[seq_record.gene_code] = []
            if seq_record.gene_code not in block_2nd:
                block_2nd[seq_record.gene_code] = []
            if seq_record.gene_code not in block_3rd:
                block_3rd[seq_record.gene_code] = []

            taxonomy_as_string = self.flatten_taxonomy(seq_record)
            taxon_id = '>{0}{1}'.format(seq_record.voucher_code,
                                        taxonomy_as_string)

            block_1st2nd[seq_record.gene_code].append('{0}\n{1}\n'.format(taxon_id,
                                                                          seq_record.first_and_second_codon_positions()))
            block_1st[seq_record.gene_code].append('{0}\n{1}\n'.format(taxon_id,
                                                                       seq_record.first_codon_position()))
            block_2nd[seq_record.gene_code].append('{0}\n{1}\n'.format(taxon_id,
                                                                       seq_record.second_codon_position()))
            block_3rd[seq_record.gene_code].append('{0}\n{1}\n'.format(taxon_id,
                                                                       seq_record.third_codon_position()))
        out = self.convert_block_dicts_to_string(block_1st2nd, block_1st, block_2nd, block_3rd)
        return out

    def convert_block_dicts_to_string(self, block_1st2nd, block_1st, block_2nd, block_3rd):
        """Takes into account whether we need to output all codon positions."""
        out = ""
        # We need 1st and 2nd positions
        if self.codon_positions in ['ALL', '1st-2nd']:
            for gene_code, seqs in block_1st2nd.items():
                out += '>{0}_1st-2nd\n----\n'.format(gene_code)
                for seq in seqs:
                    out += seq
        elif self.codon_positions == '1st':
            for gene_code, seqs in block_1st.items():
                out += '>{0}_1st\n----\n'.format(gene_code)
                for seq in seqs:
                    out += seq
        elif self.codon_positions == '2nd':
            for gene_code, seqs in block_2nd.items():
                out += '>{0}_2nd\n----\n'.format(gene_code)
                for seq in seqs:
                    out += seq

        # We also need 3rd positions
        if self.codon_positions in ['ALL', '3rd']:
            for gene_code, seqs in block_3rd.items():
                out += '\n>{0}_3rd\n----\n'.format(gene_code)
                for seq in seqs:
                    out += seq
        return out

    def make_datablock_by_gene(self, block):
        out = None
        for seq_record in block:
            if not out:
                out = '[{0}]\n'.format(seq_record.gene_code)
            taxonomy_as_string = self.flatten_taxonomy(seq_record)
            taxon_id = '{0}{1}'.format(seq_record.voucher_code,
                                       taxonomy_as_string)

            sequence = get_seq(seq_record, self.codon_positions,
                               aminoacids=self.aminoacids,
                               degenerate=self.degenerate)
            seq = sequence.seq
            if sequence.warning:
                self.warnings.append(sequence.warning)

            out += '{0}{1}\n'.format(taxon_id.ljust(55), seq)
        return out

    def flatten_taxonomy(self, seq_record):
        out = ''
        if seq_record.taxonomy is None:
            return out
        else:
            if 'family' in seq_record.taxonomy:
                out += '_' + seq_record.taxonomy['family']
            if 'genus' in seq_record.taxonomy:
                out += '_' + seq_record.taxonomy['genus']
            if 'species' in seq_record.taxonomy:
                out += '_' + seq_record.taxonomy['species']

            return re.sub('_+', '_', out)


class DatasetFooter(object):
    """Builds charset block:

    Parameters:
        data (namedtuple):      with necessary info for dataset creation.
        codon_positions (str):  `1st`, `2nd`, `3rd`, `1st-2nd`, `ALL`.
        partitioning (str):     `by gene`, `by codon position`, `1st-2nd, 3rd`.
        outgroup (str):         voucher code to be used as outgroup for NEXUS
                                and TNT files.

    Example:

        >>>
        begin mrbayes;
        charset ArgKin = 1-596;
        charset COI-begin = 597-1265;
        charset COI_end = 1266-2071;
        charset ef1a = 2072-3311;
        charset RpS2 = 3312-3722;
        charset RpS5 = 3723-4339;
        charset wingless = 4340-4739;
        set autoclose=yes;
        prset applyto=(all) ratepr=variable brlensp=unconstrained:Exp(100.0) shapepr=exp(1.0) tratiopr=beta(2.0,1.0);
        lset applyto=(all) nst=mixed rates=gamma [invgamma];
        unlink statefreq=(all);
        unlink shape=(all) revmat=(all) tratio=(all) [pinvar=(all)];
        mcmc ngen=10000000 printfreq=1000 samplefreq=1000 nchains=4 nruns=2 savebrlens=yes [temp=0.11];
         sump relburnin=yes [no] burninfrac=0.25 [2500];
         sumt relburnin=yes [no] burninfrac=0.25 [2500] contype=halfcompat [allcompat];
        END;

    """
    def __init__(self, data, codon_positions=None, partitioning=None,
                 outgroup=None):
        self.data = data
        self.codon_positions = codon_positions
        self.partitioning = partitioning
        self.outgroup = outgroup

        self._validate_partitioning(partitioning)
        self._validate_codon_positions(codon_positions)
        self.charset_block = self.make_charset_block()
        self.partition_line = self.make_partition_line()

    def _validate_partitioning(self, partitioning):
        if partitioning is None:
            self.partitioning = 'by gene'
        elif partitioning not in ['by gene', 'by codon position', '1st-2nd, 3rd']:
            raise AttributeError("Partitioning parameter should be one of these: "
                                 "None, 'by gene', 'by codon position', '1st-2nd, 3rd")

    def _validate_codon_positions(self, codon_positions):
        if codon_positions is None:
            self.codon_positions = 'ALL'
        elif codon_positions not in ['1st', '2nd', '3rd', '1st-2nd', 'ALL']:
            raise AttributeError("Codon positions parameter should be one of these: "
                                 "None, '1st', '2nd', '3rd', '1st-2nd', 'ALL'")

    def make_charset_block(self):
        """
        Override this function for Phylip dataset as the content is different and
        goes into a separate file.
        """
        out = 'begin mrbayes;\n'
        out += self.make_charsets()
        return out.strip()

    def make_charsets(self):
        """
        Override this function for Phylip dataset as the content is different and
        goes into a separate file.
        """
        count_start = 1
        out = ''
        for gene_code, lengths in self.data.gene_codes_and_lengths.items():
            count_end = lengths[0] + count_start - 1
            out += self.format_charset_line(gene_code, count_start, count_end)
            count_start = count_end + 1
        return out

    def format_charset_line(self, gene_code, count_start, count_end):
        slash_number = self.make_slash_number()
        suffixes = self.make_gene_code_suffixes()
        corrected_count = self.correct_count_using_reading_frames(gene_code, count_start, count_end)

        out = ''
        for index, val in enumerate(suffixes):
            out += '    charset {0}{1} = {2}{3};\n'.format(gene_code, suffixes[index],
                                                           corrected_count[index], slash_number)
        return out

    def make_slash_number(self):
        """
        Charset lines have \2 or \3 depending on type of partitioning and codon
        positions requested for our dataset.

        :return:
        """
        if self.partitioning == 'by codon position' and self.codon_positions == '1st-2nd':
            return '\\2'
        elif self.partitioning in ['by codon position', '1st-2nd, 3rd'] and self.codon_positions in ['ALL', None]:
            return '\\3'
        else:
            return ''

    def make_gene_code_suffixes(self):
        try:
            return self.suffix_for_one_codon_position()
        except KeyError:
            return self.suffix_for_several_codon_positions()

    def suffix_for_one_codon_position(self):
        sufixes = {
            '1st': '_pos1',
            '2nd': '_pos2',
            '3rd': '_pos3',
        }
        return [sufixes[self.codon_positions]]

    def suffix_for_several_codon_positions(self):
        if self.codon_positions == 'ALL' and self.partitioning == 'by gene':
            return ['']
        elif self.codon_positions == '1st-2nd' and self.partitioning in ['by gene', '1st-2nd, 3rd']:
            return ['_pos12']
        elif self.codon_positions == '1st-2nd' and self.partitioning == 'by codon position':
            return ['_pos1', '_pos2']

        if self.partitioning == 'by codon position':
            return ['_pos1', '_pos2', '_pos3']
        elif self.partitioning == '1st-2nd, 3rd':
            return ['_pos12', '_pos3']

    def correct_count_using_reading_frames(self, gene_code, count_start, count_end):
        reading_frame = self.data.reading_frames[gene_code]
        bp = BasePairCount(reading_frame, self.codon_positions, self.partitioning, count_start, count_end)
        return bp.get_corrected_count()

    def make_partition_line(self):
        out = 'partition GENES = {0}: '.format(len(self.data.gene_codes) * len(self.make_gene_code_suffixes()))
        out += ', '.join(self.add_suffixes_to_gene_codes())
        out += ';'
        out += '\n\nset partition = GENES;'
        return out

    def add_suffixes_to_gene_codes(self):
        """Appends pos1, pos2, etc to the gene_code if needed."""
        out = []
        for gene_code in self.data.gene_codes:
            for sufix in self.make_gene_code_suffixes():
                out.append('{0}{1}'.format(gene_code, sufix))
        return out

    def dataset_footer(self):
        return self.make_footer()

    def make_footer(self):
        outgroup = self.get_outgroup()

        footer = """{0}\n{1}

set autoclose=yes;{2}
prset applyto=(all) ratepr=variable brlensp=unconstrained:Exp(100.0) shapepr=exp(1.0) tratiopr=beta(2.0,1.0);
lset applyto=(all) nst=mixed rates=gamma [invgamma];
unlink statefreq=(all);
unlink shape=(all) revmat=(all) tratio=(all) [pinvar=(all)];
mcmc ngen=10000000 printfreq=1000 samplefreq=1000 nchains=4 nruns=2 savebrlens=yes [temp=0.11];
 sump relburnin=yes [no] burninfrac=0.25 [2500];
 sumt relburnin=yes [no] burninfrac=0.25 [2500] contype=halfcompat [allcompat];
END;
    """.format(self.charset_block, self.partition_line, outgroup)
        return footer.strip()

    def get_outgroup(self):
        """Generates the outgroup line from the voucher code specified by the
        user.
        """
        if self.outgroup is not None:
            outgroup_taxonomy = ''
            for i in self.data.seq_records:
                if self.outgroup == i.voucher_code:
                    outgroup_taxonomy = '{0}_{1}'.format(i.taxonomy['genus'],
                                                         i.taxonomy['species'])
                    break
            outgroup = '\noutgroup {0}_{1};'.format(self.outgroup,
                                                    outgroup_taxonomy)
        else:
            outgroup = ''
        return outgroup


class BasePairCount(object):
    """
    Uses reading frame info, partitioning method and number of codon positions
    to return corrected base pair count for charset lines.

    Example:

        >>> bp_count = BasePairCount(reading_frame=1, codon_positions='1st-2nd',
        ...                          partitioning='by codon position',
        ...                          count_start=100, count_end=512)
        >>> bp_count.get_corrected_count()
        [
            '100-512',
            '101-513',
        ]
    """
    def __init__(self, reading_frame=None, codon_positions=None, partitioning=None,
                 count_start=None, count_end=None):
        self._partitioning = self._set_partitioning(partitioning)
        self._codon_positions = self._set_codon_positions(codon_positions)
        self._reading_frame = self._set_reading_frame(reading_frame)
        self._count_start = self._set_count_start(count_start)
        self._count_end = self._set_count_end(count_end)

    def _set_codon_positions(self, codon_positions):
        if not codon_positions:
            raise ValueError("_codon_positions argument is needed. Can't be None")
        else:
            return codon_positions

    def _set_reading_frame(self, reading_frame):
        if not reading_frame and self._partitioning in ['by codon position', '1st-2nd, 3rd']:
            raise ValueError("_reading_frame argument is needed. Can't be None")
        else:
            return reading_frame

    def _set_partitioning(self, partitioning):
        if not partitioning:
            raise ValueError("_partitioning argument is needed. Can't be None")
        else:
            return partitioning

    def _set_count_start(self, count_start):
        if not count_start:
            raise ValueError("codon_start argument is needed. Can't be None")
        else:
            return count_start

    def _set_count_end(self, count_end):
        if not count_end:
            raise ValueError("codon_end argument is needed. Can't be None")
        else:
            return count_end

    def get_corrected_count(self):
        if self._codon_positions == '1st-2nd' and self._partitioning in ['by gene',
                                                                         'by codon position',
                                                                         '1st-2nd, 3rd']:
            return self._using_1st2nd_codons()

        if self._codon_positions == 'ALL' and self._partitioning == 'by codon position':
            return self._using_all_codons_partition_by_codon_position()

        if self._codon_positions in ['ALL', '1st', '2nd', '3rd'] and self._partitioning == 'by gene':
            return self._using_all_codons_partition_by_gene()

        if self._codon_positions in ['1st', '2nd', '3rd'] and self._partitioning in ['by codon position', '1st-2nd, 3rd']:
            return self._using_one_codon_position_partitioned_by_codon_position(self._codon_positions)

        if self._codon_positions == 'ALL' and self._partitioning == '1st-2nd, 3rd':
            return self._using_all_codons_partition_by_1st2nd_3rd()

    def _using_1st2nd_codons(self):
        return [
            '{0}-{1}'.format(self._count_start, self._count_end),
            '{0}-{1}'.format(self._count_start + 1, self._count_end),
        ]

    def _using_all_codons_partition_by_codon_position(self):
        if self._reading_frame == 1:
            return [
                '{0}-{1}'.format(self._count_start, self._count_end),
                '{0}-{1}'.format(self._count_start + 1, self._count_end),
                '{0}-{1}'.format(self._count_start + 2, self._count_end),
            ]

        elif self._reading_frame == 2:
            return [
                '{0}-{1}'.format(self._count_start + 1, self._count_end),
                '{0}-{1}'.format(self._count_start + 2, self._count_end),
                '{0}-{1}'.format(self._count_start, self._count_end),
            ]

        else:
            return [
                '{0}-{1}'.format(self._count_start + 2, self._count_end),
                '{0}-{1}'.format(self._count_start, self._count_end),
                '{0}-{1}'.format(self._count_start + 1, self._count_end),
            ]

    def _using_one_codon_position_partitioned_by_codon_position(self, position):
        return [
            '{0}-{1}'.format(self._count_start, self._count_end),
        ]

    def _using_all_codons_partition_by_gene(self):
        return [
            '{0}-{1}'.format(self._count_start, self._count_end,)
        ]

    def _using_all_codons_partition_by_1st2nd_3rd(self):
        if self._reading_frame == 1:
            return [
                '{0}-{1}\\3 {2}-{3}'.format(self._count_start, self._count_end,
                                            self._count_start + 1,
                                            self._count_end),
                '{0}-{1}'.format(self._count_start + 2, self._count_end),
            ]

        elif self._reading_frame == 2:
            return [
                '{0}-{1}\\3 {2}-{3}'.format(self._count_start + 1,
                                            self._count_end,
                                            self._count_start + 2,
                                            self._count_end),
                '{0}-{1}'.format(self._count_start, self._count_end),
            ]

        else:
            return [
                '{0}-{1}\\3 {2}-{3}'.format(self._count_start + 2,
                                            self._count_end,
                                            self._count_start,
                                            self._count_end),
                '{0}-{1}'.format(self._count_start + 1, self._count_end),
            ]



from .base_dataset import DatasetFooter


class PhylipDatasetFooter(DatasetFooter):
    def make_charset_block(self):
        """
        Overridden function for Phylip dataset as the content is different and
        goes into a separate file.
        """
        out = self.make_charsets()
        return out.strip()

    def make_charsets(self):
        """
        Overridden function for Phylip dataset as the content is different and
        goes into a separate file.
        """
        count_start = 1
        out = ''
        for gene_code, lengths in self.data.gene_codes_and_lengths.items():
            count_end = lengths[0] + count_start - 1
            formatted_line = self.format_charset_line(gene_code, count_start, count_end)
            converted_line = formatted_line.replace('    charset', 'DNA,').replace(';', '')
            out += converted_line
            count_start = count_end + 1
        return out



import os
import unittest

from dataset_creator.dataset import Dataset

from .data import test_data

FASTA_DATA_PATH = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'Fasta')


class TestGenBankFASTA(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_dataset(self):
        dataset = Dataset(test_data, format='GenBankFASTA', partitioning='by gene')
        result = dataset.dataset_str
        expected = open(os.path.join(FASTA_DATA_PATH, 'dataset_genbankfasta.fas')).read()
        self.assertEqual(expected, result)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from __future__ import absolute_import, print_function

import io
import re
import sys
from os.path import dirname
from os.path import join

from setuptools import setup


def read(*names, **kwargs):
    return io.open(
        join(dirname(__file__), *names),
        encoding=kwargs.get('encoding', 'utf8')
    ).read()


required_libs = ['seqrecord-expanded==0.2.2']
if sys.version_info < (2, 7,):
    required_libs.append('ordereddict==1.1')

setup(
    name='dataset-creator',
    version='0.3.9',
    license='BSD',
    description='Takes SeqRecordExpanded objects and creates datasets for phylogenetic software',
    long_description='%s\n%s' % (read('README.rst'), re.sub(':[a-z]+:`~?(.*?)`', r'``\1``', read('CHANGELOG.rst'))),
    author='Carlos Peña',
    author_email='mycalesis@gmail.com',
    url='https://github.com/carlosp420/dataset-creator',
    packages=['dataset_creator'],
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Utilities',
    ],
    keywords=[
        # eg: 'keyword1', 'keyword2', 'keyword3',
    ],
    install_requires=required_libs,
    extras_require={
        # eg:
        #   'rst': ['docutils>=0.11'],
        #   ':python_version=="2.6"': ['argparse'],
    },
    entry_points={
        'console_scripts': [
            'dataset_creator = dataset_creator.__main__:main',
        ]
    },
    test_suite='tests',
)
